/*
 * Palindrome Check Master - demonstrates MASTER SLAVE
 * Code-Beispiel zum Buch Patterns Kompakt, Spektrum Akademischer Verlag
 * Copyright 2012 Karl Eilebrecht
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package de.calamanari.pk.masterslave;

import java.io.File;
import java.io.IOException;
import java.text.NumberFormat;
import java.util.Locale;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.Executor;
import java.util.concurrent.Executors;
import java.util.logging.Logger;

import de.calamanari.pk.util.IndexedTextFileAccessor;

/**
 * Palindrome Check Master - the MASTER in this MASTER SLAVE example,
 * divides the task into subtasks waits for the slaves to complete the
 * sub-tasks and returns the result.
 * @author <a href="mailto:Karl.Eilebrecht(a/t)freenet.de">Karl Eilebrecht</a>
 */
public class PalindromeCheckMaster {

	
	/**
	 * logger
	 */
	public static final Logger LOGGER = Logger.getLogger(PalindromeCheckMaster.class.getName());
	
	/**
	 * Size of the partitions a slave will work on
	 */
	private final int partitionSize;
	
	/**
	 * Internally used executor to control the slaves' work
	 */
	private final Executor executor;
	
	
	/**
	 * Creates the master with some configuration information
	 * @param numberOfSlaves number of parallel workers
	 * @param partitionSize the size (characters) of a partition (pair) a slave will work in
	 */
	public PalindromeCheckMaster(int numberOfSlaves, int partitionSize) {
		this.partitionSize = partitionSize;
		this.executor = Executors.newFixedThreadPool(numberOfSlaves);
	}
	
	/**
	 * While waiting the MASTER can do other things
	 */
	private void doOtherStuff() {
		LOGGER.fine("MASTER does other stuff while waiting ...");
		try {
			Thread.sleep(1000);
		}
		catch (InterruptedException ex) {
			// ignore
		}
	}
	
	
	/**
	 * Perform a palindrome test on the given text file.
	 * @param file source file with text to be checked
	 * @param charsetName character set, i.e. "UTF-8"
	 * @return check result
	 * @throws IOException
	 * @throws InterruptedException
	 * @throws ExecutionException
	 */
	public PalindromeCheckResult performPalindromeFileTest(File file, String charsetName) throws IOException, InterruptedException, ExecutionException {
		LOGGER.fine(this.getClass().getSimpleName() + ".performPalindromeFileTest({file=" + file + ", charsetName=" + charsetName + "}) called");
		
		PalindromeCheckResult res = PalindromeCheckResult.UNKNOWN;
		IndexedTextFileAccessor.Configuration indexerConfiguration = new IndexedTextFileAccessor.Configuration();
		indexerConfiguration.maxNumberOfCharIndexEntries = 20000;
		indexerConfiguration.maxNumberOfLineIndexEntries = 1;
		LOGGER.info("Preparing (scan and create partitions for SLAVEs) ... ");
		IndexedTextFileAccessor textFileAccessor = new IndexedTextFileAccessor(file, charsetName, indexerConfiguration);
		long numberOfCharacters = textFileAccessor.getNumberOfCharacters();
		if (numberOfCharacters < 2) {
			res = PalindromeCheckResult.CONFIRMED;
		}
		else {

			long halfLen = numberOfCharacters / 2;
			
			// (1) if the number was even, say 70 characters
			//     centerOfFile=halfLen = 35 (incl.)
			//     first half == characters [0..35) and second half == characters [35..70)
			//
			// (2) if the number was odd, say 71 characters
			//     centerOfFile=halfLen = 35 (excl.)
			//     first half == characters [0..35) and second half == characters [36..71) (ignore character 35)
			

			int numberOfFullPartitions = (int)(halfLen / partitionSize);
			int remainderPartitionSize = (int)(halfLen - ((long)numberOfFullPartitions) * partitionSize);
			
			// number of partitions is the number of section pairs (left/right) for palindrome test
			// AQAABBTBCCCCXCCCCBTBBAAQA
			//  S1  S2  S3   S4  S5  S6
            //
			// 6 sections (S1-S6), 3 partitions (pairs): [S1, S6]; [S2, S5]; [S3, S4]
			// S1 will be compared character-by-character against the reversed version of S6, same for S2 and S5 as well as for S3 and S4
			
			int numberOfPartitions = numberOfFullPartitions;
			if (remainderPartitionSize > 0) {
				numberOfPartitions++; // one (shorter) remainder partition pair
			}
			
			PalindromeCheckFuture future = new PalindromeCheckFuture(numberOfPartitions);
			
			for (int i = 0; i < numberOfFullPartitions; i++) {
				long startOfLeftPartition = ((long)i) * partitionSize;
				long startOfRightPartition = numberOfCharacters - startOfLeftPartition - partitionSize;
				PalindromeCheckSlaveTask task = new PalindromeCheckSlaveTask(textFileAccessor, startOfLeftPartition, startOfRightPartition, partitionSize, future);
				executor.execute(task);
			}
			
			if (remainderPartitionSize > 0) {
				long startOfLeftPartition = halfLen - remainderPartitionSize;
				long startOfRightPartition = numberOfCharacters - startOfLeftPartition - remainderPartitionSize;
				PalindromeCheckSlaveTask task = new PalindromeCheckSlaveTask(textFileAccessor, startOfLeftPartition, startOfRightPartition, remainderPartitionSize, future);
				executor.execute(task);
			}

			LOGGER.info("Preparation completed, MASTER is waiting for SLAVE-results ...");
			
			
			boolean done = false;
			NumberFormat nf = NumberFormat.getInstance(Locale.US);
			nf.setMaximumFractionDigits(0);
			nf.setMinimumFractionDigits(0);
			do {
				LOGGER.fine("MASTER polls for result ...");
				done = future.isDone();
				if (!done) {
					LOGGER.info("" + nf.format(future.getProgressPerc()) + "% completed, MASTER is still waiting for SLAVE-results ...");
					doOtherStuff();
				}
			}
			while (!done);

			res = future.get();
			
			LOGGER.fine("Palindrome check finished!");
			
			if (res == PalindromeCheckResult.ERROR) {
				throw new ExecutionException("Unknown problem during palindrome test, see logs for details.", null);
			}
		}
		
		return res;
	}
	
	
}
